/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkSimple3DCirclesStrategy.cxx,v $

  Copyright (c) Ferenc Nasztanovics
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "vtkSimple3DCirclesStrategy.h"

#include "vtkCharArray.h"
#include "vtkDirectedGraph.h"
#include "vtkIdTypeArray.h"
#include "vtkInEdgeIterator.h"
#include "vtkMath.h"
#include "vtkPoints.h"
#include "vtkObjectFactory.h"
#include "vtkOutEdgeIterator.h"
#include "vtkSmartPointer.h"

#include <cmath>

#include <vtkstd/algorithm>

vtkStandardNewMacro(vtkSimple3DCirclesStrategy);
vtkCxxRevisionMacro(vtkSimple3DCirclesStrategy, "$Revision: 0.1 $" );

void vtkSimple3DCirclesStrategy::PrintSelf( ostream &os, vtkIndent indent )
{
	this->Superclass::PrintSelf( os, indent );
	os << indent << "Radius : " << this->Radius << endl;
	os << indent << "Height : " << this->Height << endl;
	os << indent << "Orign  : (" << this->Orign[0] << "," << this->Orign[1] << "," << this->Orign[2] << ")" << endl;
	os << indent << "Direction  : (" << this->Direction[0] << "," << this->Direction[1] << "," << this->Direction[2] << ")" << endl;
	if ( this->Method == FixedRadiusMethod )
		os << indent << "Method: fixed radius method" << endl;
	else if ( this->Method == FixedDistanceMethod )
		os << indent << "Method: fixed distance method" << endl;
	os << indent << "MarkValue: " << this->MarkedValue << endl;
	os << indent << "MarkedStartPoints : (" << this->MarkedStartPoints << ")" << endl;
}

vtkSimple3DCirclesStrategy::vtkSimple3DCirclesStrategy( void )
: Radius(1), Height(1), Method(FixedRadiusMethod), MarkedStartPoints(0), MarkedValue(0), ForceToUseUniversalStartPointsFinder(false)
{
	this->SetOrign(0,0,0);
	this->SetDirection(0,0,1);
}

vtkSimple3DCirclesStrategy::~vtkSimple3DCirclesStrategy( void )
{
	this->SetMarkedStartPoints(0);
}

void vtkSimple3DCirclesStrategy::Layout( void )
{
	if ( this->Graph == 0 )
	{
		vtkErrorMacro( << "Graph is null!" );
		return;
	}
	if ( this->Graph->GetNumberOfVertices() == 0 )
	{
		vtkDebugMacro( << "Graph is empty!" );
		return;
	}

	vtkSmartPointer<vtkDirectedGraph> target = vtkDirectedGraph::New();
	if ( ! target->CheckedShallowCopy( this->Graph ) )
	{
		vtkErrorMacro( << "Graph must be directed graph!" );
		return;
	}

	vtkSmartPointer<vtkIntArray> layers = vtkIntArray::New();
	layers->SetNumberOfValues( target->GetNumberOfVertices() );
	for ( vtkIdType i = 0; i <= layers->GetMaxId(); ++i )
		layers->SetValue(i,-1);

	vtkStartStore start_points, order_points;
	if ( this->UniversalStartPoints( target, start_points, layers ) == 0 )
	{
		vtkErrorMacro( << "There is no start point!" );
		return;
	}
	order_points = start_points;

	vtkSmartPointer<vtkIdTypeArray> order = vtkIdTypeArray::New();
	order->SetNumberOfValues( target->GetNumberOfVertices() );
	for ( vtkIdType i = 0; i <= order->GetMaxId(); ++i )
		order->SetValue(i,-1);

	int max_layer_id = this->BuildLayers( target, start_points, layers );

	this->BuildPointOrder( target, order_points, layers, order );
	
	if ( order->GetValue( order->GetMaxId() ) == -1 )
	{
		vtkErrorMacro( << "A part of the graph is not accessable!" );
		return;
	}

	int index = 0;
	int layer = 0;
	int start = 0;
	double R = this->Radius;
	double localXYZ[3], globalXYZ[3];
	double alfa = 0;
	double Dir[3];
	int ind = 0;

	alfa = vtkMath::Norm( this->Direction );
	Dir[0] = this->Direction[0] / alfa;
	Dir[1] = this->Direction[1] / alfa;
	Dir[2] = this->Direction[2] / alfa;
	alfa = 0;

	vtkSmartPointer<vtkPoints> points = vtkPoints::New();
	points->SetNumberOfPoints( target->GetNumberOfVertices() );

	while ( index <= order->GetMaxId() )
	{
		start = index;
		R = this->Radius;
		while ( index <= order->GetMaxId() )
		{
			if ( layers->GetValue( order->GetValue(index) ) == layer )
				++index;
			else
				break;
		}

		alfa = vtkMath::DoubleTwoPi() / double( index - start );

		if ( this->Method == FixedDistanceMethod )
		{
			R = double( index - start - 1 ) * this->Radius / vtkMath::DoublePi();
		}
		else if ( this->Method == FixedRadiusMethod )
		{
			if ( ( index - start ) == 1 )
				R = 0;
		}
		else
		{
			vtkErrorMacro( << "Method must be FixedRadiusMethod or FixedDistanceMethod!" );
			return;
		}

		for ( ind = start; ind < index; ++ind )
		{
			localXYZ[0] = R * cos( double(ind) * alfa );
			localXYZ[1] = R * sin( double(ind) * alfa );
			localXYZ[2] = double(max_layer_id-layer) * this->Height;
			globalXYZ[0] = this->Orign[0] + localXYZ[0];
			globalXYZ[1] = this->Orign[1] + localXYZ[1];
			globalXYZ[2] = this->Orign[2] + localXYZ[2];
			points->SetPoint( order->GetValue(ind), globalXYZ );
		}

		++layer;
	}

	this->Graph->SetPoints( points );

	return;
}

int vtkSimple3DCirclesStrategy::UniversalStartPoints( vtkDirectedGraph * input, vtkStartStore &target, vtkIntArray * layers )
{
	int number_of_startpoints = 0;

	if ( ( this->MarkedStartPoints != 0 ) && ( ! this->ForceToUseUniversalStartPointsFinder ) )
	{
		if ( this->MarkedStartPoints->GetMaxId() == layers->GetMaxId() )
		{
			for ( vtkIdType i = 0; i < input->GetNumberOfVertices(); ++i )
			{
				if ( ( this->MarkedStartPoints->GetValue(i) == this->MarkedValue ) && ( input->GetOutDegree(i) > 0 ) )
				{
					target.push_back(i);
					layers->SetValue( i, 0 );
					++number_of_startpoints;
				}
			}

			vtkDebugMacro( << "StartPoint finder: Universal start point finder was used with initial start point list." );
			return number_of_startpoints;
		}
		else
		{
			vtkErrorMacro( << "MarkedStartPoints number is NOT equal number of vertices!" );
			return 0;
		}
	}

	for ( vtkIdType i = 0; i < input->GetNumberOfVertices(); ++i )
	{
		if ( ( input->GetInDegree(i) == 0 ) && ( input->GetOutDegree(i) > 0 ) )
		{
			target.push_back(i);
			layers->SetValue( i, 0 );
			++number_of_startpoints;
		}
	}

	vtkDebugMacro( << "StartPoint finder: Universal start point finder was used." );
	return number_of_startpoints;
}

int vtkSimple3DCirclesStrategy::BuildLayers( vtkDirectedGraph * input, vtkStartStore &source, vtkIntArray * layers )
{
	vtkSmartPointer<vtkOutEdgeIterator> edge_out_iterator = vtkOutEdgeIterator::New();
	vtkSmartPointer<vtkInEdgeIterator> edge_in_iterator = vtkInEdgeIterator::New();
	vtkIdType layer = 0, flayer = 0;
	vtkOutEdgeType out_edge;
	bool HasAllInput = true;
	int step = 0;
	vtkIdType ID = 0;
	int max_layer_id = -1;

	while ( source.size() > 0 )
	{
		ID = source.front();
		source.pop_front();

		input->GetOutEdges( ID, edge_out_iterator );

		while ( edge_out_iterator->HasNext() )
		{
			out_edge = edge_out_iterator->Next();
			if ( layers->GetValue( out_edge.Target ) == -1 )
			{
				input->GetInEdges( out_edge.Target, edge_in_iterator );
				HasAllInput = true;
				layer = layers->GetValue( ID );

				while ( edge_in_iterator->HasNext() )
				{
					flayer = layers->GetValue( edge_in_iterator->Next().Source );
					if (flayer  == -1 )
					{
						HasAllInput = false;
						break;
					}
					layer = vtkstd::max( layer, flayer );
				}

				if ( HasAllInput )
				{
					source.push_back( out_edge.Target );
					layers->SetValue( out_edge.Target, layer + 1 );
					max_layer_id = vtkstd::max( max_layer_id, layer + 1 );
				}
			}
		}
		++step;
	}

	vtkDebugMacro( << "Layer building is successful." );
	return max_layer_id;
}

void vtkSimple3DCirclesStrategy::BuildPointOrder( vtkDirectedGraph * input, vtkStartStore &source, vtkIntArray * layers, vtkIdTypeArray * order )
{
	vtkSmartPointer<vtkOutEdgeIterator> edge_out_iterator = vtkOutEdgeIterator::New();
	vtkSmartPointer<vtkCharArray> mark = vtkCharArray::New();
	vtkOutEdgeType out_edge;
	int step = 0;
	int layer = 0;
	vtkIdType ID = 0;

	int max_id = order->GetMaxId();

	mark->SetNumberOfValues( input->GetNumberOfVertices() );
	for ( vtkIdType i = 0; i <= mark->GetMaxId(); ++i )
		mark->SetValue(i,0);

	while ( source.size() > 0 )
	{
		ID = source.front();
		source.pop_front();

		order->SetValue( step, ID );
		input->GetOutEdges( ID, edge_out_iterator );
		layer = layers->GetValue( ID ) + 1;

		while ( edge_out_iterator->HasNext() )
		{
			out_edge = edge_out_iterator->Next();
			if ( mark->GetValue( out_edge.Target ) == 0 )
			{
				if ( layers->GetValue( out_edge.Target ) == layer )
				{
					mark->SetValue( out_edge.Target, 1 );
					source.push_back( out_edge.Target );
				}
			}
		}

		++step;
	}

	vtkDebugMacro( << "Vertex order building is successful." );
}
